#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

from math import *
import string
import re
import os.path
import qubtree
import QUB
import qubopt
from Buttons import AddTaskButton
from Dialogs import *
import SelectCommands

def itertree(tree, name=None):
    if name != None:
        child = tree.find(name)
        next = lambda child: child.nextSameName()
    else:
        child = tree.child
        next = lambda child: child.sibling
    while not child.isNull:
        yield child
        child = next(child)

def enumtree(tree, name=None):
    if name != None:
        child = tree.find(name)
        next = lambda child: child.nextSameName()
    else:
        child = tree.child
        next = lambda child: child.sibling
    i = 0
    while not child.isNull:
        yield i, child
        child = next(child)
        i += 1

def argmax(seq):
    val = max(seq)
    for i in xrange(len(seq)):
        if val == seq[i]:
            return i, val

def ClsfMIL(data, model, pct):
    properties = QUB.Profiles['MIL'].active.clone()
    properties.appendClone(model)
    properties['Mode'].data = 'evaluate'
    properties['use segments'].data = 'separately'
    noCB = lambda x: x
    results = qubopt.MIL(properties, [data], noCB, noCB, pct)
    for seg in itertree(results['DataSet'], 'Segment'):
        N = seg['DwellCount'].data[0]
        seg['LLperN'].data = seg['LL'].data[0] / N
    return results

def ClsfMPL(data, model, pct):
    properties = QUB.Profiles['MPL'].active.clone()
    properties.appendClone(model)
    properties['Mode'].data = 'evaluate'
    properties['use segments'].data = 'separately'
    results = qubopt.MPL(properties, [data], QUB.fillData, noCB, noCB, pct)
    for seg in itertree(results['DataSet'], 'Segment'):
        N = seg.data[1] - seg.data[0] + 1
        seg['LLperN'].data = seg['LL'].data[0] / N
    return results
    

# algorithms must produce a result whose ['DataSet']['Segment']s have ['LLperN']
algorithms = {'MIL' : ClsfMIL,
              'MPL' : ClsfMPL}

def EnsureProperties():
    props = QUB.Profiles['Clsf'].active
    algo = props['algo']
    if len(algo.data) == 0:
        algo.data = algorithms.keys()[0]
    thresh = props['thresh']
    if len(thresh.data) == 0:
        thresh.data = 1.0
    return props

def Prompt():
    props = EnsureProperties()

    mdlView = QUB.ModelView
    models = [mdlView[i].getAsNode() for i in xrange(len(mdlView))]
    # workaround for older qub with bug that ignores index in ModelView[i]:
    if (len(models) < 2) or (str(models[0]['FileName'].data) == str(models[1]['FileName'].data)):
        models = []
        savedI = mdlView.currIndex
        for i in xrange(len(mdlView)):
            mdlView.currIndex = i
            models.append( mdlView.currModel.getAsNode() )
        mdlView.currIndex = savedI
    mdlNames = [os.path.split(str(mdl['FileName'].data))[1] for mdl in models]
            
    options = [DialogCheckBox(name, name, 0) for i, name in enumerate(mdlNames)]
    options.append(DialogRadioGroup('algo', 'Algorithm:', algorithms.keys(), str(props['algo'].data)))
    options.append(DialogText('thresh', 'min LL / N:', props['thresh'].data[0], acceptFloat))
    response = QUB.Dialog('Classify selections', ['OK', 'Cancel'], options)
    if response['button'] == 'OK':
        models = [mdl for i, mdl in enumerate(models) if response[mdlNames[i]]]
        algo = algorithms[response['algo']]
        threshold = response['thresh']
    else:
        models, algo, threshold = [], None, 0.0
    props['algo'].data = response['algo']
    props['thresh'].data = response['thresh']
    return models, algo, threshold
        
def Classify(props):
    datasets = QUB.getChosenIdealization()
    data = datasets.find('DataSet')
    models, algo, threshold = Prompt()
    modelNames = [os.path.split(str(model['FileName'].data))[1] for model in models]

    if (0 == len(models)) or data.find('Segment').isNull:
        return

    wizard = QUB.CreateWizard()
    wizard.caption = 'Classify'
    wizard.show()
    
    wizard.message = 'Running...'
    def pct(x):
        wizard.message = 'Running...%s%%' % x.data[0]
        return x
    
    results = [algo(data, model, pct) for model in models]
    wizard.message = 'Classifying...'
    LLperN = [ [seg['LLperN'].data[0] for seg in itertree(result['DataSet'], 'Segment')] for result in results ]
    
    outResult = qubtree.Node('Clsf Results')
    outResult['ResultType'].data = 'Clsf'
    outResult.appendClone(datasets.find('DataDescription'))
    outData = outResult.appendClone(data)
    for s, seg in enumtree(outData, 'Segment'):
        seg.remove(seg['Firsts'])
        seg.remove(seg['Lasts'])
        seg.remove(seg['Classes'])
        seg.remove(seg['Durations'])
        for m in xrange(len(modelNames)):
            seg['LLperN %s' % modelNames[m]].data = LLperN[m][s]
        LLi, LLmax = argmax([LLperN[m][s] for m in xrange(len(LLperN))])
        if LLmax < threshold:
            seg['SelectedColor'].data = 0
        else:
            seg['SelectedColor'].data = LLi + 1

    wizard.hide()
    QUB.ResultsView.addResults(outResult)

    for m in xrange(len(modelNames)):
        SelectCommands.NewSelListOfColor(m+1, modelNames[m])

AddTaskButton( 'Classify', # button text
               Classify,   # function
               'Clsf',       # short text for small buttons
               0,            # show settings dialog on left-click?
               0,            # show settings dialog on right-click?
               "Classify selections by best model likelihood",   # tool-tip
               "Classify selections by best model likelihood",   # description in "Edit Buttons"
               "Properties:", # caption for the settings dialog
               [],
               'Modeling')

